C     Last change:  HO   10 Oct 2005    8:26 pm
c Program RESIDUW  (Version of RESIDUAL with extended capabilities)
c*****************************************************************
c
c       created: 3/5/91, last updated: 7/20/92 (SRC) and 6/3/94 (NERI)
c       Author:
c                        Joseph Chang
C                        Sigma Research Corp.
C                        234 Littleton Rd., Suite 2E
C                        Westford, MA 01886
C                        Telephone: (508) 692-0330
C
C       Additions by Helge R. Olesen              / September 2005
c                    National Environmental Research Institute
c                    P.O. Box 358
c                    Denmark
c                    e-mail hro@dmu.dk
c
c  Purpose:
c  --------
c
c       This is a preprocessor of a residual plotting program, SIGPLOT
c       It basically reads:
c       1) multiple columns of concurrent observed and model
c          predicted parameters (such as concentration) and other
c          independent variables (such as wind speed, stability, hour
c          of day), and
c       2) information used to define the ranges of each of the
c          independent variables,
c       then calculates the distribution of the ratio of any two models
c       for all ranges of each of the independent variables.
c
c  Note:
c  -----
c       As input the current programme accepts files prepared for BOOT.
c
c
c  Extended capabilities  (Helge R. Olesen)
c  ----------------------
c
c       The following capabilities have been added to the
c       original version:
c
c       A filter can be imposed, so the ratio of c_mod/c_obs is
c       assumed to be 1 when both are small.
c
c       An alternative choice of percentiles for box plots is
c       offered.
c
c       A file suitable for producing quantile-quantile plots etc.
c       can be generated
c
c
c       Changes September 2005:
C -----------------------------
c       The format of the input file is changed compared to the original, 
c       because the BOOT software requires a new format (an extra column has
c       been inserted as the first).
cs 

c  Variables and Parameters:
c  -------------------------
c
C     maxmm:     Maximum no. of models allowed.
C     maxnn:     Maximum no. of data points for each model allowed.
C     maxvar:    Maximum no. of independent variables, based on which
C                the models will be evaluated, allowed.
C     maxrng:    Maximum no. of ranges in any one of the independent
C                variables allowed.
c     c:         columns of observed and model predicted parameters
c     x:         columns of concurrent independent variables
c     ixr:       no. of ranges for each of the independent variables
c     xr:        boundaries defining ranges for each of the indep. var.
c     res:       the ratio of any two columns of c
c     wk:        working array for res
c     icount:    no. of res data points in each range of each indep. var.
c     ycum:      the CDF of res in each range of each indep. var.
c     modnam:    name of each model
c     varnam:    name of each indep. var.
c*****************************************************************

        parameter (maxmm=16, maxnn=1300, maxvar=10, maxrng=20)
        real c(maxnn,0:maxmm), x(maxnn,maxvar), wk(maxnn), res(maxnn)
        real xr(maxrng+1,maxvar)
        real ycum(5,maxrng,maxvar)
        integer icount(maxrng,maxvar), icum(maxvar), ixr(maxvar)
        integer len(0:maxmm)
        character*40 modnam(0:maxmm),varnam(maxvar)
        character*40 fname,ylabel
        character*1 iansl,iansu
        logical lansl,lansu


c >>> ADDITIONAL DECLARATIONS /  HRO:
        logical exten, filter, quant
        character*1 ians
        character*40 quanam, labeli, labelj
        real col1(maxnn), col2(maxnn)


c  Initialization
        do 10 i=1,maxnn
        c(i,0)=1.
10      continue

         quant = .false.
         filter = .false.


c
c       Opening statements.
c
        print *,'Name of input file:     '
        read *,fname
        open (unit=15,file=fname,status='old')
c
        print *,'Name of output file:    '
        read *,fname
        open (unit=16,file=fname)



C
C   nn = number of observations
C   mm = number of models (including the observed data set as one)
C   kk = number of blocks, NOT USED IN THIS PROGRAM
C   nvar = number of independent variables
C
      read (15,*) nn,mm,kk,nvar
      if (nn.gt.maxnn) stop 'MAXNN exceeded'
      if (mm.gt.maxmm) stop 'MAXMM exceeded'
      if (nvar.gt.maxvar) stop 'MAXVAR exceeded'
C
C   Skip the next line containing the blocking information that is
C   not used in this program
C
      read (15,*)

c
c   Read the model names
c
      modnam(0)='1'
      read (15,*) (modnam(i),i=1,mm)

C
C   Skip the next line containing names of each block that are
C   not used in this program
C
      read (15,*)

c
c   Find out the lenght of each MODNAM.
c
      maxlen=1
      do 50 i=0,mm
        call clen (len(i),8,modnam(i))
        maxlen=max0(maxlen,len(i))
50    continue

c
c   Pick two models whose ratio will be analyzed
c
        write (*,11)
11      format(
     1' Following models are available, choose two models, i and j,',/,
     2' where the ratio of Model(i)/Model(j) will be analyzed against',/,
     3' the independent variables:')
        do 60 i=1,mm
        print *,i,')  ',modnam(i)(1:len(i))
60      continue
        print *,0,')  ','All 1''s'
        write (*,21) mm
21      format(' Enter the two numbers, in order, between 0 and ',i2)
        print *
        read *,ii,jj

c
c  Set lower and upper threshold values on the ratio,
c  this is sometimes necessary if the box plot will be drawn on
c  a logarithmic scale
c
        print *,'Implement a lower threshold on the ratio (<Y>/n) ?   '
        read '(a1)',iansl
        lansl=.false.
        if (iansl.eq.' ' .or. iansl.eq.'y' .or. iansl.eq.'Y') then
        lansl=.true.
        print *,'Enter the lower threshold on the ratio (e.g. 0.01)   '
        read *,epsl
        end if

        print *,'Implement a upper threshold on the ratio (<Y>/n) ?   '
        read '(a1)',iansu
        lansu=.false.
        if (iansu.eq.' ' .or. iansu.eq.'y' .or. iansu.eq.'Y') then
        lansu=.true.
        print *,'Enter the upper threshold on the ratio (e.g. 100.)   '
        read *,epsu
        end if

c >>> ADDITION /  HRO:

c
c      Sequence of additional questions.
c      These questions are not asked if "ratio" has 1 in the denominator
c      or in the numerator.
c
       i_box=1
       fillim = 0.
       if (ii .eq. 0 .or. jj .eq. 0) goto 70

c
c      Use of extended capabilities of RESIDUW?
c
       print *,
     & 'Do you wish to use extended capabilities of RESIDUW (<Y>/n) ? '
        read '(a1)',ians
        exten=.false.
        if (ians.eq.' ' .or. ians.eq.'y' .or. ians.eq.'Y') then
           exten=.true.

c
c      Filter?
c
           print *, ' '
           print *, ' '
           print *, ' '
           print *,
     &'Do you wish to impose a filter, so that the ratio ',
     &  modnam(ii)(1:len(ii)), '/', modnam(jj)(1:len(jj)),' is'
           print *,
     &'considered unity when both ', modnam(ii)(1:len(ii)), ' and ',
     & modnam(jj)(1:len(jj)), ' are small ?  '
           print *, '    '
           print *,
     &'(This will remove noise, thus making it easier to     '
           print *,
     &'diagnose tendencies in data)                                  '
           print *, '                                        (<Y>/n)  '
           read '(a1)',ians
           filter=.false.
           if (ians.eq.' ' .or. ians.eq.'y' .or. ians.eq.'Y') then
              filter=.true.

              print *,
     &'What is the threshold for this filter (e.g. 15.) ?   '
              read *,fillim
c                                                           (end filter)
           endif


c
c      Alternative set of percentiles?
c

           print 700
700   format(
     &'     '////
     &' Normally, on box plots produced by RESIDUAL, the following   '/
     &' percentiles are indicated:                                   '/
     &' (1)      2.4    16       50       84   and   97.6            '/
     &'                                                              '/
     &' For plots of ratios, this is not always pertinent. Often, the'/
     &' boxes stretch over many decades, so that no tendencies can be'/
     &' seen in such plots.                                          '/
     &'                                                             '/
     &' Therefore, you can choose to use an alternative set of       '/
     &' percentiles:                                                 '/
     &' (2)      5      25      50       75          95             '/
     &'                                                             '/
     &' Enter 1 or 2, according to your desire >>>>   ')

80         continue
           read *, i_box
           if (i_box .lt. 1 .or. i_box .gt. 2) go to 80
c
c      File for Quantile-Quantile plot?
c
           print  701
701   format(
     &'                                                               '/
     &' Do you want to create a file suitable for a                   '/
     &' quantile-quantile-plot (<Y>/n) ?   ')
           read '(a1)',ians
           quant=.false.
           if (ians.eq.' ' .or. ians.eq.'y' .or. ians.eq.'Y') then
              quant=.true.
c
        print *,
     &    'Name of additional output file (suggested type: .QUA):  '
              read *,quanam
              open (unit=17,file=quanam)
c                               (end of quantile-quantile questionning)
           endif


c                                      (end of non-standard section)
       endif
C  >>> End of additional questions
70     continue

C
C  Read observed and modelled parameters (e.g. concentrations, dosages),
C  plus the concurrent independent variables.
C
      print *,'Reading data'
      do 100 n=1,nn
      read (15,*) dummy, (c(n,m),m=1,mm),(x(n,m),m=1,nvar)
100   continue


c
c  Read the range information of each independent variable
c  ixr:  no. of ranges of the ith variable
c  xr: bounds used to define the ixr ranges.
c
        do 120 i=1,nvar
        read (15,*,end=900) ixr(i),varnam(i),(xr(j,i),j=1,ixr(i)+1)
        if (ixr(i).gt.maxrng) stop 'MAXRNG exceeded'
120     continue

c
c  Define the ratio (residual) of Model(ii)/Model(jj)
c
        i_res1=0
        do 140 i=1,nn
        if (filter .and. c(i,jj).lt. fillim .and. c(i,ii) .lt. fillim)
     &  then
          res(i)=1.
          i_res1=i_res1+1
        else if (c(i,jj).eq.0.) then
          if (lansu) then
            res(i)=epsu
          else
            print *,
     1      'ERROR found, a zero value is found in the denominator'
            stop
          end if
        else
          res(i)=c(i,ii)/c(i,jj)
        end if
140     continue

c
c  Implement lower threshold if necessary.
c
        if (lansl) then
        do 150 i=1,nn
        res(i)=amax1(res(i),epsl)
150     continue
        end if


c
c  Analyze RES against all independent variables.
c
        do 160 k=1,nvar
        call strtfy (maxnn,maxrng,res,x(1,k),nn,ixr(k),xr(1,k),
     1  icount(1,k),ycum(1,1,k),icum(k),wk, i_box)
160     continue

c
c  Generate output ready to be plotted by the BOX5 program:
c
        if (jj.eq.0) then
          ylabel=modnam(ii)
        else
          ylabel=modnam(ii)(1:len(ii))//'/'//modnam(jj)(1:len(jj))
        end if
        do 200 k=1,nvar
        if (i_box .eq. 1) write (16,*) ' '
        if (i_box .eq. 2) write (16,*) 'Box percentiles: 5, 25, etc'
        write (16,*) varnam(k)
        write (16,*) ylabel
        write (16,*) icum(k),' 1'
        do 210 j=1,ixr(k)
        if (icount(j,k).le.4) go to 210
        write (16,111) (xr(j,k)+xr(j+1,k))/2., (ycum(i,j,k),i=1,5),
     *  icount(j,k)
111     format(6f10.4,i6)
210     continue
200     continue


c
c  Generate output for quantile-quantile plots etc.
c
      if (quant) then
         labeli=modnam(ii)(1:len(ii))
         labelj=modnam(jj)(1:len(jj))

         if (filter) then
         write (17, 702) fillim
702      format (' Each column is sorted. A filter value of ',
     &     f12.4, ' has been used for ratio.')
         else
            write (17, *)
     &      ' Each column is sorted. No filter applied to ratio.'
         end if
         write (17,703) labelj
         write (17,703) labeli
703      format (1x,a8)
         write (17, 705) nn, labelj, labeli, ylabel
705      format (1x,i4, '  1  ', a8,4x,a8,3x,a17)
         write (17,* )     ''' '''
         do 220 iii=1,nn
            col1(iii)=c(iii,jj)
            col2(iii)=c(iii,ii)
220      continue
         call sort('a',nn,col1)
         call sort('a',nn,col2)
         call sort('a',nn,res)
         do 230 iii=1,nn
            write (17, 710) col1(iii), col2(iii), res(iii)
710         format (4x,3f12.4)
230      continue
      end if

c
c  Summary of choices
c
      print *,'  '
      print *,'  '
      print *,'  '
      print *,'  '
      print *,
     &'Summary of choices made:                                       '
      if (lansl) print *,
     &'A lower threshold of ',epsl, ' has been imposed on the ratio,'

      if (lansu) print *,
     &'An upper threshold of ',epsu, ' has been imposed on the ratio.'

      if (filter) then
         print *,
     &   'A filter has been imposed on small values of variables.'
         write (*,715) fillim, i_res1, nn
715      format (' The filter value is ', f8.2  /
     &          ' The ratio has been changed to unity in ',i4, ' of ',
     &          i4, ' cases.')
      endif

      if (i_box .eq. 2) print *,
     &'The alternative set of percentiles (5, 25, 50, 75, 95) is used.'

      call clen (lfname,40,fname)
      print *,
     &'The file ', fname(1:lfname), ' has been created.      '

      if (quant) call clen (lqname,40,quanam)
      if (quant) print *,
     &'The file ', quanam(1:lqname),
     &' has been created as basis for q-q plots.'

        goto 999
900     print *, 'Error in an attempt to read bottom lines in file'
        print *,
     &   '(where ranges of independent variables should be defined).'

999     stop
        end


c======================================================================
c   Subroutine strtfy
c
c   Purpose:
c   --------
c
c   Determines the distribution of the dependent variable, Y, on
c   different ranges of the independent variable, X.  Within each
c   range of X, the CDF of Y will be determined.
c
c   Arguments:
c   ----------
c
c   input:
c   Y      real     Array of size N containing the dependent variable
c                   to be analyzed.
c   X      real     Array of size N containing the independent variable,
c                   concurrent with Y, according to which Y will be
c                   analyzed.
c   N      integer  Dimension of X and Y.
c   MAXNN  integer  Maximum size of N.
c   IXR    integer  No. of ranges in X to be considered.
c   MAXRNG integer  Maximum size of IXR.
c   XR     real     Array of size IXR+1 defining the boundaries of the
c                   ranges in X.
c
c   output:
c   ICOUNT integer  Array of size IXR storing the number of Y data points
c                   in each range of X.
c   YCUM   real     Array of size (5,IXR) containing the CDF at 2.3,
c                   15.9, 50, 84.1, and 97.7 percentiles of Y for each
c                   range of x.  YCUM for a certain range, say j, in X
c                   will only be calculated if ICOUNT(j) is at least 5.
c   ICUM   integer  Number of ranges in X where there are at least 5
c                   Y data points present.
c   WK     real     Working array of size N used to store all Y data
c                   points in each range of X.
c   I_CDF  integer  1 for the usual set of CDF percentiles (2.3 etc.);
c                   2 for an alternative set of CDF percentiles:
c                   5, 25, 50, 75 and 95 percentiles.
c
c======================================================================
       subroutine strtfy
     &   (maxnn,maxrng,y,x,n,ixr,xr,icount,ycum,icum,wk, i_cdf)
       real x(maxnn),y(maxnn),ycum(5,maxrng),xr(maxrng+1),wk(maxnn)
       integer icount(maxrng)
c
c       Initialization
c
        icum=0
        do 10 i=1,ixr
        icount(i)=0
        do 10 j=1,5
        ycum(j,i)=0.
10      continue
c
c       Loop all data over all ranges of y.
c
        do 40 j=1,ixr
        do 20 i=1,n
        if (xr(j).le.x(i) .and. x(i).lt.xr(j+1)) then
          icount(j)=icount(j)+1
          wk(icount(j))=y(i)
        end if
20      continue
c
c       Start sorting and calculating CDF for each range of x
c
        if (icount(j).lt.5) goto 40
        icum=icum+1
        call sort('a',icount(j),wk)
        call stat(wk,icount(j),avg,sd,ycum(1,j), i_cdf)
40      continue
        return
        end


c
c======================================================================
c   Subroutine stat
c
c   Purpose:
c   To calcualte the CDF of an array
c======================================================================
      subroutine stat(a,n,avg,sd,cum, i_cdf)
        real a(n)
        real cum(5)
c
c  According to I_CDF, one of two alternative sets of percentiles is
c  computed
c           i_cdf=1                i_cdf=2
c  cum(1):    2.3%                   4.9%
c  cum(2):    15.9%                  24.4%
c  cum(3):    50%                    50%
c  cum(4):    84.1%                  75.1%
c  cum(5):    97.7%                  95.1%
c
        if (n.lt.5) then
          print *,'Error in STAT: N is < 5'
          stop
        end if
c
        sum1=0.
        sum2=0.
        do 10 i=1,n
        sum1=sum1+a(i)
        sum2=sum2+a(i)*a(i)
10      continue
        avg=sum1/float(n)
        xx=sum2/float(n) - avg*avg
        if (xx.le.0.) then
        sd=0.
        else
        sd=sqrt(xx)
        end if
c
        if (i_cdf .eq. 1) then
           if (n.gt.5) then
             j=ifix(.023*float(n))+1
             cum(1)=a(j)
             j=ifix(.159*float(n))+1
             cum(2)=a(j)
             j=ifix(.50*float(n))+1
             cum(3)=a(j)
             j=ifix(.841*float(n))+1
             cum(4)=a(j)
             j=ifix(.977*float(n))+1
             cum(5)=a(j)
           else
             cum(1)=a(1)
             cum(2)=a(2)
             cum(3)=a(3)
             cum(4)=a(4)
             cum(5)=a(5)
           end if
        else if (i_cdf .eq. 2) then
           if (n.gt.5) then
             j=ifix(.049*float(n))+1
             cum(1)=a(j)
             j=ifix(.249*float(n))+1
             cum(2)=a(j)
             j=ifix(.50*float(n))+1
             cum(3)=a(j)
             j=ifix(.751*float(n))+1
             cum(4)=a(j)
             j=ifix(.951*float(n))+1
             cum(5)=a(j)
           else
             cum(1)=a(1)
             cum(2)=a(2)
             cum(3)=a(3)
             cum(4)=a(4)
             cum(5)=a(5)
           end if
         end if
c
        return
        end


c======================================================================
c     subroutine sort
c
c     created:7/22/88, last updated:7/22/88
c     this routine does the sorting for vector a of size n
c     according to the shell-metzner scheme.
c
c     note that in the calling program of sort, a may be declared
c     with a larger size.
c
c     the order of sorting, ascending or descending, is controlled
c     by the switch isw,
c     isw = 'a' or 'A' ---->  ascending.
c     isw = 'd' or 'D' ----> descending.
c
c======================================================================
      subroutine sort (isw, n, a)
      integer n, i, j, jump, j2, j3
      character*1 isw
      real a(n)
c
      jump = n
c
      if (isw. eq. 'a' .or. isw.eq.'A') goto 2000
      if (isw. eq. 'd' .or. isw.eq.'D') goto 4000
      print *,'Check the first parameter of SORT in the calling program.
     *'
      print *,'It has to be ''a'', ''A'', ''d'', or ''D''.'
      stop
c
c --- section 1.  in ascending order.
c
2000  continue
10    jump = jump / 2
      if (jump .eq. 0) goto 99
      j2 = n - jump
      do 30 j = 1, j2
15    i = j
20    j3 = i + jump
      if (a(i). le. a(j3)) goto 30
      call swap (a(i), a(j3))
      i = i - jump
      if (i. gt. 0) goto 20
30    continue
      go to 10
99    return
c
c --- section 2.  in descending order.
c
4000  continue
210    jump = jump / 2
      if (jump .eq. 0) goto 299
      j2 = n - jump
      do 230 j = 1, j2
215   i = j
220   j3 = i + jump
      if (a(i). ge. a(j3)) goto 230
      call swap (a(i), a(j3))
      i = i - jump
      if (i. gt. 0) goto 220
230   continue
      go to 210
299   return
      end
c
c======================================================================
c     subroutine swap
c======================================================================
      subroutine swap (a, b)
      real a, b, hold
      hold = a
      a = b
      b = hold
      return
      end
c======================================================================
c  subroutine clen
c
c  Find the length of a character string "a"
c  Note that the leading blanks will also be counted as part
c  of the character string.
c
c  len:   the length of a, stripping off all trailing blanks
c  lena:  the original lenght of a declared in the calling program
c
c======================================================================
        subroutine clen(len,lena,a)
        character *(*) a
        do 10 i=lena,1,-1
        if (a(i:i).ne.' ') goto 11
10      continue
        print *,'Warning in CLEN, "a" is all blanks!'
        len=0
        return
11      len=i
        return
        end
